@@ -27,7 +27,9 @@
  Oct4.2002 : AS : corrected bug in heap allocation, that caused a crash during sequencer free.
 */
 
-#if !C_FLUIDSYNTH && defined(WIN32)
+#define FLUIDINC
+#include "config.h"
+#if !C_FLUIDSYNTH && defined(WIN32) && !defined(C_HX_DOS)
 #include "fluid_event_priv.h"
 #include "fluidsynth_priv.h"
 
