@@ -169,7 +169,7 @@
 
 #endif
 
-#if defined(_WIN32) && !defined(C_HX_DOS)
+#if defined(_WIN32) && !defined(C_HX_DOS) && !defined(_WIN32_WINDOWS)
 #define flac_stat_s __stat64 /* stat struct */
 #define flac_fstat _fstat64
 #else
