@@ -49,6 +49,7 @@
 #include <dirent.h>
 #endif
 #include "build_timestamp.h"
+#include "version_string.h"
 
 extern bool shell_keyboard_flush;
 extern bool dos_shell_running_program, mountwarning, winautorun;
@@ -84,7 +85,7 @@ bool isDBCSCP(), InitCodePage(), isKanji1(uint8_t chr), shiftjis_lead_byte(int c
 Bitu call_shellstop = 0;
 /* Larger scope so shell_del autoexec can use it to
  * remove things from the environment */
-DOS_Shell * first_shell = 0; 
+DOS_Shell * first_shell = nullptr;
 
 static Bitu shellstop_handler(void) {
 	return CBRET_STOP;
@@ -234,7 +235,7 @@ void AutoexecObject::Uninstall() {
 			char* buf2 = new char[n + 1];
 			safe_strncpy(buf2, buf.c_str(), n + 1);
 			bool stringset = false;
-			// If it's a environment variable remove it from there as well
+			// If it's an environment variable remove it from there as well
 			if ((strncasecmp(buf2,"set ",4) == 0) && (strlen(buf2) > 4)){
 				char* after_set = buf2 + 4;//move to variable that is being set
 				char* test2 = strpbrk(after_set,"=");
@@ -274,7 +275,7 @@ DOS_Shell::DOS_Shell():Program(){
 	echo=true;
 	exit=false;
 	perm = false;
-	bf=0;
+	bf = nullptr;
 	call=false;
 	exec=false;
 	lfnfor = uselfn;
@@ -381,7 +382,7 @@ public:
     std::string str = "";
     unsigned long long curpos = 0;
 	device_TMP(char *name) { SetName(name); };
-	virtual bool Read(uint8_t * data,uint16_t * size) {
+	bool Read(uint8_t * data,uint16_t * size) override {
         int i;
         for (i=0; i<*size; i++) {
             if (curpos+i>=str.size()) break;
@@ -390,11 +391,11 @@ public:
 		*size = i;
 		return true;
 	}
-	virtual bool Write(const uint8_t * data,uint16_t * size) {
+	bool Write(const uint8_t * data,uint16_t * size) override {
         for (int i=0; i<*size; i++) str += std::string(1, data[i]);
 		return true;
 	}
-	virtual bool Seek(uint32_t * pos,uint32_t type) {
+	bool Seek(uint32_t * pos,uint32_t type) override {
 		switch (type) {
             case 0:
                 curpos = *pos;
@@ -410,13 +411,12 @@ public:
                 return false;
 		}
 		if (curpos > str.size()) curpos = str.size();
-		else if (curpos < 0) curpos = 0;
 		return true;
 	}
-	virtual bool Close() { return true; }
-	virtual uint16_t GetInformation(void) { return (strcmp(RunningProgram, "WCLIP") ? DeviceInfoFlags::Device : 0) | DeviceInfoFlags::EofOnInput; }
-	virtual bool ReadFromControlChannel(PhysPt bufptr,uint16_t size,uint16_t * retcode) { (void)bufptr; (void)size; (void)retcode; return false; }
-	virtual bool WriteToControlChannel(PhysPt bufptr,uint16_t size,uint16_t * retcode) { (void)bufptr; (void)size; (void)retcode; return false; }
+	bool Close() override { return true; }
+	uint16_t GetInformation(void) override { return (strcmp(RunningProgram, "WCLIP") ? DeviceInfoFlags::Device : 0) | DeviceInfoFlags::EofOnInput; }
+	bool ReadFromControlChannel(PhysPt bufptr,uint16_t size,uint16_t * retcode) override { (void)bufptr; (void)size; (void)retcode; return false; }
+	bool WriteToControlChannel(PhysPt bufptr,uint16_t size,uint16_t * retcode) override { (void)bufptr; (void)size; (void)retcode; return false; }
 };
 
 void DOS_Shell::ParseLine(char * line) {
@@ -426,10 +426,10 @@ void DOS_Shell::ParseLine(char * line) {
 	line = trim(line);
 
 	/* Do redirection and pipe checks */
-	
-	char * in  = 0;
-	char * out = 0;
-	char * toc = 0;
+
+	char * in  = nullptr;
+	char * out = nullptr;
+	char * toc = nullptr;
 
 	uint16_t dummy,dummy2;
 	uint32_t bigdummy = 0;
@@ -450,7 +450,7 @@ void DOS_Shell::ParseLine(char * line) {
 			DOS_OpenFile(in,OPEN_READ,&dummy);	//Open new stdin
 		} else {
 			WriteOut(!*in?"File open error\n":(dos.errorcode==DOSERR_ACCESS_DENIED?MSG_Get("SHELL_CMD_FILE_ACCESS_DENIED"):"File open error - %s\n"), in);
-			in = 0;
+			in = nullptr;
 			return;
 		}
 	}
@@ -754,7 +754,7 @@ void GetExpandedPath(std::string &path) {
 void showWelcome(Program *shell) {
     /* Start a normal shell and check for a first command init */
     ansiinstalled = is_ANSI_installed(shell);
-    std::string verstr = "v"+std::string(VERSION)+", "+GetPlatform(false);
+    std::string verstr = "v"+std::string(VERSION)+", "+OS_PLATFORM_LONG+" "+SDL_STRING+" " + OS_BIT + "-bit";
     if (machine == MCH_PC98) {
         shell->WriteOut(ParseMsg("\x86\x52\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44"
             "\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44"
@@ -762,7 +762,7 @@ void showWelcome(Program *shell) {
             "\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44"
             "\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44"
             "\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x44\x86\x56\n"));
-        shell->WriteOut(ParseMsg((std::string("\x86\x46 \033[32m")+(MSG_Get("SHELL_STARTUP_TITLE")+std::string("             ")).substr(0,30)+std::string("  \033[33m%*s\033[37m \x86\x46\n")).c_str()),34,verstr.c_str());
+        shell->WriteOut(ParseMsg((std::string("\x86\x46 \033[32m")+(MSG_Get("SHELL_STARTUP_TITLE")+std::string("      ")).substr(0,22)+std::string("  \033[33m%*s\033[37m \x86\x46\n")).c_str()),42,verstr.c_str());
         shell->WriteOut(ParseMsg("\x86\x46                                                                    \x86\x46\n"));
         shell->WriteOut(ParseMsg((std::string("\x86\x46 ")+MSG_Get("SHELL_STARTUP_HEAD1_PC98")+std::string(" \x86\x46\n")).c_str()));
         shell->WriteOut(ParseMsg("\x86\x46                                                                    \x86\x46\n"));
@@ -891,7 +891,8 @@ void DOS_Shell::Prepare(void) {
                     } else if (control->opt_langcp && !name && (layout.empty() || layout=="auto"))
                         SetKEYBCP();
                 }
-                if (lastmsgcp && lastmsgcp != dos.loaded_codepage) SwitchLanguage(lastmsgcp, dos.loaded_codepage, true);
+                //if (lastmsgcp && lastmsgcp != dos.loaded_codepage) SwitchLanguage(lastmsgcp, dos.loaded_codepage, true);
+                if (msgcodepage && msgcodepage != dos.loaded_codepage) SwitchLanguage(dos.loaded_codepage, msgcodepage, true);
             }
 			if (country>0&&!control->opt_noconfig) {
 				countryNo = country;
@@ -1092,38 +1093,38 @@ public:
         int cp=dos.loaded_codepage;
         InitCodePage();
         force_conversion = false;
-        int ind=0;
 
         /* The user may have given .BAT files to run on the command line */
         if (!control->auto_bat_additional.empty()) {
             std::string cmd = "@echo off\n";
 
-            for (unsigned int i=0;i<control->auto_bat_additional.size();i++) {
+            decltype(control->auto_bat_additional)::size_type ind = 0;
+            for (auto & str : control->auto_bat_additional) {
                 if (!control->opt_prerun) cmd += "\n";
-                if (!strncmp(control->auto_bat_additional[i].c_str(), "@mount c: ", 10)) {
-                    cmd += control->auto_bat_additional[i]+"\n";
+                if (!strncmp(str.c_str(), "@mount c: ", 10)) {
+                    cmd += str+"\n";
                     cmd += "@config -get lastmount>nul\n";
                     cmd += "@if not '%CONFIG%'=='' %CONFIG%";
                 } else {
                     std::string batname;
-                    //LOG_MSG("auto_bat_additional %s\n", control->auto_bat_additional[i].c_str());
+                    //LOG_MSG("auto_bat_additional %s\n", str.c_str());
 
-                    std::replace(control->auto_bat_additional[i].begin(),control->auto_bat_additional[i].end(),'/','\\');
+                    std::replace(str.begin(),str.end(),'/','\\');
                     size_t pos = std::string::npos;
                     bool lead = false;
-                    for (unsigned int j=0; j<control->auto_bat_additional[i].size(); j++) {
+                    for (unsigned int j=0; j<str.size(); j++) {
                         if (lead) lead = false;
-                        else if ((IS_PC98_ARCH && shiftjis_lead_byte(control->auto_bat_additional[i][j])) || (isDBCSCP() && isKanji1(control->auto_bat_additional[i][j]))) lead = true;
-                        else if (control->auto_bat_additional[i][j]=='\\') pos = j;
+                        else if ((IS_PC98_ARCH && shiftjis_lead_byte(str[j])) || (isDBCSCP() && isKanji1(str[j]))) lead = true;
+                        else if (str[j]=='\\') pos = j;
                     }
                     if(pos == std::string::npos) {  //Only a filename, mount current directory
-                        batname = control->auto_bat_additional[i];
+                        batname = str;
                         cmd += "@mount c: . -nl -q\n";
                     } else { //Parse the path of .BAT file
-                        std::string batpath = control->auto_bat_additional[i].substr(0,pos+1);
+                        std::string batpath = str.substr(0,pos+1);
                         if (batpath==".\\") batpath=".";
                         else if (batpath=="..\\") batpath="..";
-                        batname = control->auto_bat_additional[i].substr(pos+1);
+                        batname = str.substr(pos+1);
                         cmd += "@mount c: \"" + batpath + "\" -nl -q\n";
                     }
                     std::string opt = control->opt_o.size() > ind && control->opt_o[ind].size() ? " "+control->opt_o[ind] : "";
@@ -1163,7 +1164,7 @@ public:
 
 		if (control->opt_prerun) RunAdditional();
 
-		/* add stuff from the configfile unless -noautexec or -securemode is specified. */
+		/* add stuff from the configfile unless -noautoexec or -securemode is specified. */
 		const char * extra = section->data.c_str();
 		if (extra && !secure && !control->opt_noautoexec) {
 			/* detect if "echo off" is the first line */
@@ -1239,7 +1240,7 @@ public:
 				if (access(buffer,F_OK)) continue;
 				autoexec[12].Install(std::string("MOUNT C \"") + buffer + "\"");
 				autoexec[13].Install("C:");
-				/* Save the non-modified filename (so boot and imgmount can use it (long filenames, case sensivitive)) */
+				/* Save the non-modified filename (so boot and imgmount can use it (long filenames, case sensitive)) */
 				strcpy(orig,name);
 				upcase(name);
 				if(strstr(name,".BAT") != 0) {
@@ -1296,7 +1297,7 @@ static void AUTOEXEC_ShutDown(Section * sec) {
 	}
     if (first_shell != NULL) {
 		delete first_shell;
-		first_shell = 0;//Make clear that it shouldn't be used anymore
+		first_shell = nullptr;//Make clear that it shouldn't be used anymore
     }
     if (call_shellstop != 0) {
         CALLBACK_DeAllocate(call_shellstop);
@@ -1415,10 +1416,13 @@ void SHELL_Init() {
 	MSG_Add("SHELL_CMD_MKDIR_EXIST","Directory already exists - %s\n");
 	MSG_Add("SHELL_CMD_MKDIR_ERROR","Unable to create directory - %s\n");
 	MSG_Add("SHELL_CMD_RMDIR_ERROR","Invalid path, not directory, or directory not empty - %s\n");
+    MSG_Add("SHELL_CMD_RMDIR_FULLTREE_CONFIRM", "Delete directory \"%s\" and all its subdirectories? (Y/N)?");
+    MSG_Add("SHELL_CMD_RMDIR_SINGLE_CONFIRM", "Delete file \"%s\" (Y/N)?");
     MSG_Add("SHELL_CMD_RENAME_ERROR","Unable to rename - %s\n");
 	MSG_Add("SHELL_CMD_ATTRIB_GET_ERROR","Unable to get attributes: %s\n");
 	MSG_Add("SHELL_CMD_ATTRIB_SET_ERROR","Unable to set attributes: %s\n");
 	MSG_Add("SHELL_CMD_DEL_ERROR","Unable to delete - %s\n");
+    MSG_Add("SHELL_CMD_DEL_CONFIRM", "Delete %s (Y/N)?");
 	MSG_Add("SHELL_CMD_DEL_SURE","All files in directory will be deleted!\nAre you sure [Y/N]?");
 	MSG_Add("SHELL_SYNTAXERROR","Syntax error\n");
 	MSG_Add("SHELL_CMD_SET_NOT_SET","Environment variable %s not defined.\n");
@@ -1450,6 +1454,7 @@ void SHELL_Init() {
 	MSG_Add("SHELL_CMD_COPY_CONFIRM","Overwrite %s (Yes/No/All)?");
 	MSG_Add("SHELL_CMD_COPY_NOSPACE","Insufficient disk space - %s\n");
 	MSG_Add("SHELL_CMD_COPY_ERROR","Copy error - %s\n");
+    MSG_Add("SHELL_CMD_COPY_NOSELF", "File cannot be copied onto itself\r\n");
 	MSG_Add("SHELL_CMD_SUBST_DRIVE_LIST","The currently mounted local drives are:\n");
 	MSG_Add("SHELL_CMD_SUBST_NO_REMOVE","Unable to remove, drive not in use.\n");
 	MSG_Add("SHELL_CMD_SUBST_IN_USE","Target drive is already in use.\n");
@@ -1458,6 +1463,7 @@ void SHELL_Init() {
 	MSG_Add("SHELL_CMD_SUBST_FAILURE","SUBST: There is an error in your command line.\n");
 	MSG_Add("SHELL_CMD_VTEXT_ON","DOS/V V-text is currently enabled.\n");
 	MSG_Add("SHELL_CMD_VTEXT_OFF","DOS/V V-text is currently disabled.\n");
+    MSG_Add("SHELL_ALLFILES_CHAR", "a");
 
     std::string mapper_keybind = mapper_event_keybind_string("host");
     if (mapper_keybind.empty()) mapper_keybind = "unbound";
@@ -1730,7 +1736,7 @@ void SHELL_Init() {
 	MSG_Add("SHELL_CMD_TRUENAME_HELP","Finds the fully-expanded name for a file.\n");
 	MSG_Add("SHELL_CMD_TRUENAME_HELP_LONG","TRUENAME [/H] file\n");
 	MSG_Add("SHELL_CMD_DXCAPTURE_HELP","Runs program with video or audio capture.\n");
-	MSG_Add("SHELL_CMD_DXCAPTURE_HELP_LONG","DX-CAPTURE [/V|/-V] [/A|/-A] [/M|/-M] [command] [options]\n\nIt will start video or audio capture, run program, and then automatically stop capture when the program exits.\n");
+	MSG_Add("SHELL_CMD_DXCAPTURE_HELP_LONG","DX-CAPTURE [/V|/-V] [/A|/-A] [/M|/-M] [/O|/-O] [command] [options]\n\nIt will start video or audio capture, run program, and then automatically stop capture when the program exits.\n /V for video, /A for audio, /M multi-track audio and /O for OPL FM (DROv2 format)");
 #if C_DEBUG
 	MSG_Add("SHELL_CMD_DEBUGBOX_HELP","Runs program and breaks into debugger at entry point.\n");
 	MSG_Add("SHELL_CMD_DEBUGBOX_HELP_LONG","DEBUGBOX [command] [options]\n\nType DEBUGBOX without a parameter to start the debugger.\n");
@@ -1759,12 +1765,25 @@ void SHELL_Init() {
 	uint16_t env_seg;//=DOS_FIRST_SHELL+19; //DOS_GetMemory(1+(4096/16))+1;
 	uint16_t stack_seg;//=DOS_GetMemory(2048/16,"COMMAND.COM stack");
     uint16_t tmp,total_sz;
+    bool tiny_memory_mode = false;
+
+    // below a certain memory size, alter memory arrangement and allocation to minimize memory
+    if (MEM_TotalPages() < 0x8) tiny_memory_mode = true;
 
     // decide shell env size
-    if (dosbox_shell_env_size == 0)
-        dosbox_shell_env_size = (0x158u - (0x118u + 19u)) << 4u; /* equivalent to mainline DOSBox */
-    else
+    if (dosbox_shell_env_size == 0) {
+        if (MEM_TotalPages() >= 0x10/*64KB or more*/)
+            dosbox_shell_env_size = (0x158u - (0x118u + 19u)) << 4u; /* equivalent to DOSBox SVN */
+        else if (MEM_TotalPages() >= 0x8/*32KB or more*/)
+            dosbox_shell_env_size = 384;
+        else if (MEM_TotalPages() >= 0x4/*16KB or more*/)
+            dosbox_shell_env_size = 256;
+        else
+            dosbox_shell_env_size = 144;
+    }
+    else {
         dosbox_shell_env_size = (dosbox_shell_env_size+15u)&(~15u); /* round up to paragraph */
+    }
 
     LOG_MSG("COMMAND.COM env size:             %u bytes",dosbox_shell_env_size);
 
@@ -1787,9 +1806,20 @@ void SHELL_Init() {
     LOG_MSG("COMMAND.COM environment block:    0x%04x sz=0x%04x",env_seg,tmp);
 
     // COMMAND.COM main binary (including PSP and stack)
-    tmp = 0x1A + (2048/16);
+    if (tiny_memory_mode)
+        tmp = 0x13 + (1536/16);
+    else
+        tmp = 0x1A + (2048/16);
     total_sz = tmp;
-	if (!DOS_AllocateMemory(&psp_seg,&tmp)) E_Exit("COMMAND.COM failed to allocate main body + PSP segment");
+
+    // Use normal MCB allocation unless memsize is 4KB
+    if (MEM_TotalPages() > 1) {
+        if (!DOS_AllocateMemory(&psp_seg,&tmp)) E_Exit("COMMAND.COM failed to allocate main body + PSP segment");
+    }
+    else {
+        psp_seg = DOS_GetMemory(tmp,"COMMAND.COM main body and PSP");
+    }
+
     LOG_MSG("COMMAND.COM main body (PSP):      0x%04x sz=0x%04x",psp_seg,tmp);
 
     DOS_SetMemAllocStrategy(savedMemAllocStrategy);
@@ -1811,12 +1841,21 @@ void SHELL_Init() {
     }
 
     // set the stack at 0x1A
-    stack_seg = psp_seg + 0x1A;
+    if (tiny_memory_mode)
+        stack_seg = psp_seg + 0x13;
+    else
+        stack_seg = psp_seg + 0x1A;
     LOG_MSG("COMMAND.COM stack:                0x%04x",stack_seg);
 
     // set the stack pointer
 	SegSet16(ss,stack_seg);
-	reg_sp=2046;
+
+    if (tiny_memory_mode)
+        reg_sp=1534;
+    else
+        reg_sp=2046;
+
+    LOG(LOG_MISC,LOG_DEBUG)("Shell init SS:SP %04x:%04x",(unsigned int)stack_seg,(unsigned int)reg_sp);
 
 	/* Set up int 24 and psp (Telarium games) */
 	real_writeb(psp_seg+16+1,0,0xea);		/* far jmp */
@@ -1964,7 +2003,7 @@ void SHELL_Run() {
 	try {
 		first_shell->Run();
 		delete first_shell;
-		first_shell = 0;//Make clear that it shouldn't be used anymore
+		first_shell = nullptr;//Make clear that it shouldn't be used anymore
 		prepared = false;
 		dos_shell_running_program = false;
 #if DOSBOXMENU_TYPE == DOSBOXMENU_HMENU
@@ -1973,7 +2012,7 @@ void SHELL_Run() {
 	}
 	catch (...) {
 		delete first_shell;
-		first_shell = 0;//Make clear that it shouldn't be used anymore
+		first_shell = nullptr;//Make clear that it shouldn't be used anymore
 		prepared = false;
 		dos_shell_running_program = false;
 #if DOSBOXMENU_TYPE == DOSBOXMENU_HMENU
